module Graph
       (Vertex(..), vertices, dualV, Edge(..), formatE, domE, codomE,
        outgoingE, incomingE, initialE, terminalE, dualE)
       where
import Prelude
       (Bounded, Eq, Ord, Show, ShowS, (/=), maxBound, minBound)
import Data.Ix (Ix, range)
 
data Vertex = V0
            | VA
            | VB
            | VC
            | VD
            | VE
            | VF
            | VG
            | VH
            | VI
            | VJ
            | VK
            | VL
            deriving (Bounded, Eq, Ix, Ord, Show)
 
vertices :: [Vertex]
vertices = [v | v <- range (minBound, maxBound), V0 /= v]
 
dualV :: Vertex -> Vertex
dualV v
  = case v of
        V0 -> V0
        VA -> VL
        VB -> VK
        VC -> VJ
        VD -> VI
        VE -> VH
        VF -> VG
        VG -> VF
        VH -> VE
        VI -> VD
        VJ -> VC
        VK -> VB
        VL -> VA
 
data Edge = E0 !Vertex !Vertex
          | Ealpha
          | Ebeta
          | Egamma
          | Edelta
          | Eepsilon
          | Ezeta
          | Eeta
          | Etheta
          | Elambda
          | Emu
          | Enu
          | Exi
          | Epi
          | Erho
          | Esigma
          | Etau
          | Ephi
          | Eomega
          deriving (Eq, Ord, Show)
 
formatE :: Edge -> ShowS
formatE e
  = ((case e of
          E0 _v _cov -> '0'
          Ealpha -> '\945'
          Ebeta -> '\946'
          Egamma -> '\947'
          Edelta -> '\948'
          Eepsilon -> '\949'
          Ezeta -> '\950'
          Eeta -> '\951'
          Etheta -> '\952'
          Elambda -> '\955'
          Emu -> '\956'
          Enu -> '\957'
          Exi -> '\958'
          Epi -> '\960'
          Erho -> '\961'
          Esigma -> '\963'
          Etau -> '\964'
          Ephi -> '\966'
          Eomega -> '\969') :)
 
domE :: Edge -> Vertex
domE e
  = case e of
        E0 v _cov -> v
        Ealpha -> VA
        Ebeta -> VB
        Egamma -> VA
        Edelta -> VB
        Eepsilon -> VC
        Ezeta -> VD
        Eeta -> VE
        Etheta -> VD
        Elambda -> VE
        Emu -> VF
        Enu -> VG
        Exi -> VH
        Epi -> VG
        Erho -> VH
        Esigma -> VI
        Etau -> VJ
        Ephi -> VK
        Eomega -> VC
 
codomE :: Edge -> Vertex
codomE e
  = case e of
        E0 _v cov -> cov
        Ealpha -> VB
        Ebeta -> VC
        Egamma -> VD
        Edelta -> VE
        Eepsilon -> VF
        Ezeta -> VE
        Eeta -> VF
        Etheta -> VG
        Elambda -> VH
        Emu -> VI
        Enu -> VH
        Exi -> VI
        Epi -> VJ
        Erho -> VK
        Esigma -> VL
        Etau -> VK
        Ephi -> VL
        Eomega -> VJ
 
outgoingE :: Vertex -> [Edge]
outgoingE v
  = case v of
        V0 -> []
        VA -> [Egamma]
        VB -> [Edelta]
        VC -> [Eepsilon]
        VD -> [Ezeta, Etheta]
        VE -> [Eeta, Elambda]
        VF -> [Emu]
        VG -> [Enu, Epi]
        VH -> [Exi, Erho]
        VI -> [Esigma]
        VJ -> []
        VK -> []
        VL -> []
 
incomingE :: Vertex -> [Edge]
incomingE v
  = case v of
        V0 -> []
        VA -> []
        VB -> []
        VC -> []
        VD -> [Egamma]
        VE -> [Edelta, Ezeta]
        VF -> [Eepsilon, Eeta]
        VG -> [Etheta]
        VH -> [Elambda, Enu]
        VI -> [Emu, Exi]
        VJ -> [Epi]
        VK -> [Erho]
        VL -> [Esigma]
 
initialE :: Vertex -> Edge
initialE = E0 V0
 
terminalE :: Vertex -> Edge
terminalE v = E0 v V0
 
dualE :: Edge -> Edge
dualE e
  = case e of
        E0 v cov -> E0 (dualV cov) (dualV v)
        Ealpha -> Ephi
        Ebeta -> Etau
        Egamma -> Esigma
        Edelta -> Erho
        Eepsilon -> Epi
        Ezeta -> Exi
        Eeta -> Enu
        Etheta -> Emu
        Elambda -> Elambda
        Emu -> Etheta
        Enu -> Eeta
        Exi -> Ezeta
        Epi -> Eepsilon
        Erho -> Edelta
        Esigma -> Egamma
        Etau -> Ebeta
        Ephi -> Ealpha
        Eomega -> Eomega
