module Main (main) where
import Prelude
       (Ord, Show, IO, Int, Ordering, String, ($), (+), (++), (.), (/=),
        (<=), compare, concatMap, fst, id, mapM_, putStrLn, return)
import Data.List (sortBy)
import Data.Maybe (fromMaybe, listToMaybe, maybeToList)
import qualified Brain as B
import Graph (Vertex(..))
import qualified Map as M
import Path
       (dom, codom, outgoing, incoming, inv, comp, kernelInv, image,
        unKernelInv, unImage)
import Proof
       (Proof(..), formatProof, conclusion, invProof, dualProof,
        complexityProof)
import Prop
       (Prop(..), formatProp, domProp, invProp, dualProp, complexityProp)
import qualified Queue as Q
import Snake (axioms, hypotheses, goals)
 
data Agenda = A (Q.Queue Int Proof) (M.Map Prop Int)
            deriving Show
 
hypothesize :: Prop -> [Proof]
hypothesize prop
  = [Hypothesized (invProp prop) | V0 /= domProp prop] ++
      [Hypothesized prop]
 
insert :: Proof -> Agenda -> Agenda
insert proof a@(A q m)
  = let prop = conclusion proof in
      if 6 <= complexityProp prop then a else
        let c = complexityProof proof
            c' = fromMaybe (1 + c) (M.lookup prop m)
          in
          if c' <= c then a else A (Q.insert c proof q) (M.insert prop c m)
 
insertAll :: [Proof] -> Agenda -> Agenda
insertAll proofs a
  = case proofs of
        [] -> a
        proof : proofs' -> insertAll proofs' (insert proof a)
 
monotoneConsequences :: Proof -> [Proof]
monotoneConsequences proof
  = let SubsetOf p1 p2 = conclusion proof in
      [Monotone proof (SubsetOf (f p1) (f p2)) | V0 /= dom p1,
       f <- [kernelInv, image]]
        ++
        [Monotone proof (SubsetOf (comp p1 p3) (comp p2 p3)) |
         p3 <- incoming (dom p1)]
        ++
        [Monotone proof (SubsetOf (comp p0 p1) (comp p0 p2)) |
         p0 <- outgoing (codom p1)]
 
transitiveConsequences :: Proof -> B.Brain -> [Proof]
transitiveConsequences proof12 brain
  = let SubsetOf p1 p2 = conclusion proof12 in
      [Transitive proof12 proof23 (SubsetOf p1 p3) |
       (p3, proof23) <- B.findLower p2 brain]
        ++
        [Transitive proof01 proof12 (SubsetOf p0 p2) |
         (p0, proof01) <- B.findUpper p1 brain]
 
universalConsequences :: Proof -> [Proof]
universalConsequences proof
  = let SubsetOf p1 p2 = conclusion proof in
      [Universal proof (f (SubsetOf (comp (comp p1' (inv p1')) p2') p2'))
       | p1' <- maybeToList (unKernelInv p1),
       p2' <- maybeToList (unKernelInv p2), f <- [id, invProp]]
        ++
        [Universal proof (f (SubsetOf p1' (comp (comp p2' (inv p2')) p1')))
         | p1' <- maybeToList (unImage p1), p2' <- maybeToList (unImage p2),
         f <- [id, invProp]]
 
consequences :: Proof -> B.Brain -> [Proof]
consequences proof brain
  = monotoneConsequences proof ++ transitiveConsequences proof brain
      ++ universalConsequences proof
 
pump :: Agenda -> B.Brain -> B.Brain
pump (A q m) brain
  = fromMaybe brain $
      do (proof, q') <- Q.deleteMin q
         let a' = A q' m
         return $
           if B.member (conclusion proof) brain then pump a' brain else
             pump (insertAll (consequences proof brain) a')
               (B.insert proof brain)
 
comparing :: (Ord k) => (v -> k) -> v -> v -> Ordering
comparing f v1 v2 = compare (f v1) (f v2)
 
retrieveFrom :: B.Brain -> Prop -> String
retrieveFrom brain goal
  = fromMaybe
      ("\9760\9760\9760 " ++ formatProp goal " \9760\9760\9760\n\n\n")
      $
      listToMaybe
        [formatProof (f proof) "\n\n" |
         (prop, f) <- sortBy (comparing fst)
                        [(goal, id), (invProp goal, invProof), (dualProp goal, dualProof),
                         (dualProp (invProp goal), dualProof . invProof)],
         proof <- maybeToList (B.lookup prop brain)]
 
main :: IO ()
main
  = let brain
          = pump
              (insertAll (concatMap hypothesize (axioms ++ hypotheses))
                 (A Q.empty M.empty))
              B.empty
      in mapM_ (putStrLn . retrieveFrom brain) goals
