module Proof
       (Proof(..), formatProof, conclusion, invProof, dualProof,
        complexityProof)
       where
import Prelude
       (Show, Int, ShowS, (+), (++), (-), (.), fst, max, quot)
import Pic
       (Pic, empty, up, right, over, formatPic, picAndWidthFromString,
        horizontalRule)
import Prop (Prop, formatProp, invProp, dualProp)
 
default (Int)
 
data Tree = Node [Tree] Prop
          deriving Show
 
picAndWidthFromTrees :: [Tree] -> (Pic, Int)
picAndWidthFromTrees prems
  = let dxGap = 3 in
      case prems of
          [] -> (empty, - dxGap)
          prem : prems'
            -> let (picPrem, dxPrem) = picAndWidthFromTree prem
                   (picPrems', dxPrems') = picAndWidthFromTrees prems'
                 in
                 (picPrem `over` right (dxPrem + dxGap) picPrems',
                  dxPrem + dxGap + dxPrems')
 
picAndWidthFromTree :: Tree -> (Pic, Int)
picAndWidthFromTree (Node prems conc)
  = let (picPrems, dxPrems) = picAndWidthFromTrees prems
        (picConc, dxConc) = picAndWidthFromString (formatProp conc "")
        dx = max dxPrems dxConc
      in
      (up 2 (right ((dx - dxPrems) `quot` 2) picPrems) `over`
         up 1 (horizontalRule dx)
         `over` right ((dx - dxConc) `quot` 2) picConc,
       dx)
 
formatTree :: Tree -> ShowS
formatTree = formatPic . fst . picAndWidthFromTree
 
data Proof = Hypothesized !Prop
           | Monotone !Proof !Prop
           | Transitive !Proof !Proof !Prop
           | Universal !Proof !Prop
           deriving Show
 
stripMonotone :: Proof -> Proof
stripMonotone proof
  = case proof of
        Monotone proof' _prop -> stripMonotone proof'
        _proof -> proof
 
treesFromProof :: Proof -> [Tree]
treesFromProof proof
  = case proof of
        Transitive proof'1 proof'2 _prop
          -> treesFromProof proof'1 ++ treesFromProof proof'2
        _proof -> [treeFromProof proof]
 
treeFromProof :: Proof -> Tree
treeFromProof proof
  = case proof of
        Hypothesized prop -> Node [] prop
        Monotone proof' prop
          -> Node [treeFromProof (stripMonotone proof')] prop
        Transitive proof'1 proof'2 prop
          -> Node (treesFromProof proof'1 ++ treesFromProof proof'2) prop
        Universal proof' prop -> Node [treeFromProof proof'] prop
 
formatProof :: Proof -> ShowS
formatProof = formatTree . treeFromProof
 
conclusion :: Proof -> Prop
conclusion proof
  = case proof of
        Hypothesized prop -> prop
        Monotone _proof' prop -> prop
        Transitive _proof'1 _proof'2 prop -> prop
        Universal _proof' prop -> prop
 
invProof :: Proof -> Proof
invProof proof
  = case proof of
        Hypothesized prop -> Hypothesized (invProp prop)
        Monotone proof' prop -> Monotone (invProof proof') (invProp prop)
        Transitive proof'1 proof'2 prop
          -> Transitive (invProof proof'1) (invProof proof'2) (invProp prop)
        Universal proof' prop -> Universal (invProof proof') (invProp prop)
 
dualProof :: Proof -> Proof
dualProof proof
  = case proof of
        Hypothesized prop -> Hypothesized (dualProp prop)
        Monotone proof' prop -> Monotone (dualProof proof') (dualProp prop)
        Transitive proof'1 proof'2 prop
          -> Transitive (dualProof proof'2) (dualProof proof'1)
               (dualProp prop)
        Universal proof' prop
          -> Universal (dualProof proof') (dualProp prop)
 
complexityProof :: Proof -> Int
complexityProof proof
  = 1 +
      case proof of
          Hypothesized _prop -> 0
          Monotone proof' _prop -> complexityProof proof'
          Transitive proof'1 proof'2 _prop
            -> complexityProof proof'1 + complexityProof proof'2
          Universal proof' _prop -> complexityProof proof'
